# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import enum
import random

import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.types import CryptoDataEnumCodedBase
from cryptodatahub.tls.algorithm import TlsGreaseOneByte, TlsGreaseTwoByte

from cryptoparser.common.base import ParsableBase
from cryptoparser.common.parse import ParserBinary, ComposerBinary


class TlsInvalidType(enum.IntEnum):
    GREASE = 0
    UNKNOWN = 1


@attr.s
class TlsInvalidTypeParamsBase():
    code = attr.ib(validator=attr.validators.instance_of(int))
    value_type = attr.ib(validator=attr.validators.in_(TlsInvalidType))

    @classmethod
    def get_code_size(cls):
        raise NotImplementedError()


class TlsInvalidTypeParamsOneByte(TlsInvalidTypeParamsBase):
    @classmethod
    def get_code_size(cls):
        return 1


class TlsInvalidTypeParamsTwoByte(TlsInvalidTypeParamsBase):
    @classmethod
    def get_code_size(cls):
        return 2


@attr.s
class TlsInvalidTypeBase(ParsableBase):
    code = attr.ib(validator=attr.validators.instance_of((int, CryptoDataEnumCodedBase)))
    value = attr.ib(init=False, validator=attr.validators.instance_of(TlsInvalidTypeParamsBase))

    def __attrs_post_init__(self):
        if isinstance(self.code, self.get_grease_enum()):
            value_type = TlsInvalidType.GREASE
            self.code = self.code.value.code
        elif isinstance(self.code, CryptoDataEnumCodedBase):
            value_type = TlsInvalidType.UNKNOWN
            self.code = self.code.value.code
        else:
            try:
                self.code = self.get_grease_enum().from_code(self.code).value.code
                value_type = TlsInvalidType.GREASE
            except InvalidValue:
                value_type = TlsInvalidType.UNKNOWN
        self.value = self.get_param_class()(self.code, value_type)

    @classmethod
    @abc.abstractmethod
    def get_param_class(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_grease_enum(cls):
        raise NotImplementedError()

    @classmethod
    def get_byte_num(cls):
        return cls.get_param_class().get_code_size()

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)
        parser.parse_numeric('code', cls.get_byte_num())

        code = parser['code']
        return cls(code), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.code, self.get_byte_num())

        return composer.composed_bytes

    @classmethod
    def from_random(cls):
        grease_enum_type = cls.get_grease_enum()

        return cls(random.choice(list(grease_enum_type)))


class TlsInvalidTypeOneByte(TlsInvalidTypeBase):
    @classmethod
    def get_param_class(cls):
        return TlsInvalidTypeParamsOneByte

    @classmethod
    def get_grease_enum(cls):
        return TlsGreaseOneByte


class TlsInvalidTypeTwoByte(TlsInvalidTypeBase):
    @classmethod
    def get_param_class(cls):
        return TlsInvalidTypeParamsTwoByte

    @classmethod
    def get_grease_enum(cls):
        return TlsGreaseTwoByte



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

import datetime

from collections import OrderedDict

import attr
import urllib3

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import InvalidType
from cryptoparser.common.field import (
    FieldValueDateTime,
    FieldValueString,
    NameValuePair,
    NameValuePairListSemicolonSeparated,
)

from .classes import (
    ComponentStringEnumTest,
    FieldValueJsonTest,
    FieldValueMultipleTest,
    FieldValueMultipleExtendableTest,
    FieldValueEnumTest,
    FieldValueStringEnumTest,
    FieldValueComponentBoolTest,
    FieldValueComponentFloatTest,
    FieldValueComponentNumberTest,
    FieldValueComponentOptionTest,
    FieldValueComponentQuotedStringTest,
    FieldValueComponentPercentTest,
    FieldValueComponentStringEnumTest,
    FieldValueComponentStringTest,
    FieldValueComponentTimeDeltaTest,
    FieldValueComponentUrlTest,
    FieldValueTimeDeltaTest,
)


class TestFieldValueString(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(FieldValueString.parse_exact_size(b'value').value, 'value')

    def test_compose(self):
        self.assertEqual(FieldValueString('value').compose(), b'value')

    def test_convert(self):
        self.assertEqual(FieldValueString.convert(None), None)
        self.assertEqual(FieldValueString.convert(bytearray(b'non-string-value')), bytearray(b'non-string-value'))
        self.assertEqual(FieldValueString.convert('value'), FieldValueString('value'))

    def test_markdown(self):
        self.assertEqual(FieldValueString('value').as_markdown(), 'value')


class TestFieldValueStringEnum(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueStringEnumTest('value not in enum')
        self.assertEqual(context_manager.exception.value, 'value not in enum')

        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueStringEnumTest.parse_exact_size(b'value not in enum')
        self.assertEqual(context_manager.exception.value, 'value not in enum')

    def test_parse(self):
        self.assertEqual(
            FieldValueStringEnumTest.parse_exact_size(b'first'),
            FieldValueStringEnumTest(FieldValueEnumTest.FIRST)
        )
        self.assertEqual(
            FieldValueStringEnumTest.parse_exact_size(b'FIRST'),
            FieldValueStringEnumTest(FieldValueEnumTest.FIRST)
        )

    def test_compose(self):
        self.assertEqual(
            FieldValueStringEnumTest(FieldValueEnumTest.SECOND).compose(),
            b'second'
        )

    def test_markdown(self):
        self.assertEqual(
            FieldValueStringEnumTest(FieldValueEnumTest.FIRST).as_markdown(),
            'FiRsT'
        )
        self.assertEqual(
            FieldValueEnumTest.FIRST.value.as_markdown(),
            'FiRsT'
        )

        self.assertEqual(
            FieldValueStringEnumTest(FieldValueEnumTest.SECOND).as_markdown(),
            'second'
        )
        self.assertEqual(
            FieldValueEnumTest.SECOND.value.as_markdown(),
            'second'
        )


class TestFieldValueComponentOption(unittest.TestCase):
    def test_parse(self):
        component, _ = FieldValueComponentOptionTest.parse_immutable(b'option')
        self.assertEqual(component.value, False)

        component, _ = FieldValueComponentOptionTest.parse_immutable(b'testOption')
        self.assertEqual(component.value, True)

    def test_compose(self):
        self.assertEqual(FieldValueComponentOptionTest(False).compose(), b'')
        self.assertEqual(FieldValueComponentOptionTest(True).compose(), b'testOption')

    def test_as_markdown(self):
        self.assertEqual(FieldValueComponentOptionTest(False).as_markdown(), 'no')
        self.assertEqual(FieldValueComponentOptionTest(True).as_markdown(), 'yes')


class TestFieldValueComponentString(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidType):
            FieldValueComponentStringTest.parse_exact_size(b'shortNam')

        with self.assertRaises(InvalidType):
            FieldValueComponentStringTest.parse_exact_size(b'wrongName=value')

    def test_parse(self):
        component = FieldValueComponentStringTest.parse_exact_size(b'testString=value')
        self.assertEqual(component.value, 'value')

    def test_compose(self):
        self.assertEqual(FieldValueComponentStringTest('value').compose(), b'testString=value')

    def test_as_markdown(self):
        self.assertEqual(FieldValueComponentStringTest('value').as_markdown(), 'value')


class TestFieldValueComponentUrl(unittest.TestCase):
    _component_url_https = FieldValueComponentUrlTest('https://example.com')
    _component_url_https_bytes = b'testUrl=https://example.com'
    _component_url_mailto = FieldValueComponentUrlTest('mailto:user@example.com')
    _component_url_mailto_bytes = b'testUrl=mailto:user@example.com'

    def test_error_invalid_value(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentUrlTest.parse_exact_size(b'testUrl=https://example.com:port')
        self.assertEqual(context_manager.exception.value, 'https://example.com:port')

        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentUrlTest(None)
        self.assertEqual(context_manager.exception.value, None)

    def test_parse(self):
        component = FieldValueComponentUrlTest.parse_exact_size(b'testUrl=https://example.com')
        self.assertEqual(component.value, urllib3.util.parse_url('https://example.com'))

        component = FieldValueComponentUrlTest.parse_exact_size(b'testUrl=mailto:user@example.com')
        self.assertEqual(component.value, urllib3.util.parse_url('mailto:user@example.com'))

    def test_compose(self):
        self.assertEqual(
            FieldValueComponentUrlTest('https://example.com').compose(),
            b'testUrl=https://example.com'
        )
        self.assertEqual(
            FieldValueComponentUrlTest('mailto:user@example.com').compose(),
            b'testUrl=mailto:user@example.com'
        )

    def test_as_markdown(self):
        self.assertEqual(self._component_url_https.as_markdown(), 'https://example.com')
        self.assertEqual(self._component_url_mailto.as_markdown(), 'mailto:user@example.com')


class TestFieldValueComponentStringEnum(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentStringEnumTest.parse_exact_size(  # pylint: disable=expression-not-assigned
                b'testStringEnum=non-existing-value'
            )
        self.assertEqual(context_manager.exception.value, 'non-existing-value')

        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentStringEnumTest('non-existing-value')
        self.assertEqual(context_manager.exception.value, 'non-existing-value')

    def test_parse(self):
        self.assertEqual(
            FieldValueComponentStringEnumTest.parse_exact_size(b'testStringEnum=one'),
            FieldValueComponentStringEnumTest(ComponentStringEnumTest.ONE)
        )

    def test_compose(self):
        self.assertEqual(
            FieldValueComponentStringEnumTest(ComponentStringEnumTest.TWO).compose(),
            b'testStringEnum=two'
        )


class TestFieldValueComponentQuotedString(unittest.TestCase):
    def test_error(self):
        component = FieldValueComponentQuotedStringTest.parse_exact_size(b'testQuotedString="value')
        self.assertEqual(component.value, 'value')

        component = FieldValueComponentQuotedStringTest.parse_exact_size(b'testQuotedString=value"')
        self.assertEqual(component.value, 'value')

        component = FieldValueComponentQuotedStringTest.parse_exact_size(b'testQuotedString=""value"')
        self.assertEqual(component.value, 'value')

        component = FieldValueComponentQuotedStringTest.parse_exact_size(b'testQuotedString="value""')
        self.assertEqual(component.value, 'value')

    def test_parse(self):
        component = FieldValueComponentQuotedStringTest.parse_exact_size(b'testQuotedString=value')
        self.assertEqual(component.value, 'value')

        component = FieldValueComponentQuotedStringTest.parse_exact_size(b'testQuotedString="value"')
        self.assertEqual(component.value, 'value')

    def test_compose(self):
        self.assertEqual(FieldValueComponentQuotedStringTest('value').compose(), b'testQuotedString="value"')

    def test_markdown(self):
        self.assertEqual(FieldValueComponentQuotedStringTest('value').as_markdown(), 'value')


class TestFieldValueComponentBool(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentBoolTest.parse_exact_size(b'testBool=str')
        self.assertEqual(context_manager.exception.value, b'str')

    def test_parse(self):
        component = FieldValueComponentBoolTest.parse_exact_size(b'testBool=yes')
        self.assertTrue(component.value)

        component = FieldValueComponentBoolTest.parse_exact_size(b'testBool=no')
        self.assertFalse(component.value)

    def test_compose(self):
        self.assertEqual(FieldValueComponentBoolTest(True).compose(), b'testBool=yes')
        self.assertEqual(FieldValueComponentBoolTest(False).compose(), b'testBool=no')

    def test_as_markdown(self):
        self.assertEqual(FieldValueComponentBoolTest(True).as_markdown(), 'yes')
        self.assertEqual(FieldValueComponentBoolTest(False).as_markdown(), 'no')


class TestFieldValueComponentFloat(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentFloatTest.parse_exact_size(b'testFloat=str')
        self.assertEqual(context_manager.exception.value, b'str')

    def test_parse(self):
        component = FieldValueComponentFloatTest.parse_exact_size(b'testFloat=1')
        self.assertEqual(component.value, 1.0)

        component = FieldValueComponentFloatTest.parse_exact_size(b'testFloat=1.0')
        self.assertEqual(component.value, 1.0)

    def test_compose(self):
        self.assertEqual(FieldValueComponentFloatTest(1).compose(), b'testFloat=1.0')
        self.assertEqual(FieldValueComponentFloatTest(1.0).compose(), b'testFloat=1.0')

    def test_as_markdown(self):
        self.assertEqual(FieldValueComponentFloatTest(1).as_markdown(), '1.0')
        self.assertEqual(FieldValueComponentFloatTest(1.0).as_markdown(), '1.0')


class TestFieldValueComponentNumber(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentNumberTest.parse_exact_size(b'testNumber=notnumeric')
        self.assertEqual(context_manager.exception.value, b'notnumeric')

    def test_parse(self):
        component = FieldValueComponentNumberTest.parse_exact_size(b'testNumber=1234')
        self.assertEqual(component.value, 1234)

    def test_compose(self):
        self.assertEqual(FieldValueComponentNumberTest(1234).compose(), b'testNumber=1234')

    def test_as_markdown(self):
        self.assertEqual(FieldValueComponentNumberTest(1234).as_markdown(), '1234')


class TestFieldValueComponentPercent(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentPercentTest.parse_exact_size(b'testPercent=101')
        self.assertEqual(context_manager.exception.value, 101)


class TestFieldValueComponentTimeDelta(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentTimeDeltaTest.parse_exact_size(b'testTimeDelta=notnumeric')
        self.assertEqual(context_manager.exception.value, b'notnumeric')

        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueComponentTimeDeltaTest.parse_exact_size(
                b'testTimeDelta=' + str(2 ** 48).encode('ascii')
            )
        self.assertEqual(context_manager.exception.value, 2 ** 48)

    def test_parse(self):
        component = FieldValueComponentTimeDeltaTest.parse_exact_size(b'testTimeDelta=86401')
        self.assertEqual(component.value, datetime.timedelta(days=1, seconds=1))

    def test_compose(self):
        self.assertEqual(
            FieldValueComponentTimeDeltaTest(datetime.timedelta(days=1, seconds=1)).compose(),
            b'testTimeDelta=86401'
        )

    def test_as_markdown(self):
        self.assertEqual(
            FieldValueComponentTimeDeltaTest(datetime.timedelta(days=1, seconds=1)).as_markdown(),
            '1 day, 0:00:01'
        )


class TestFieldJson(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueJsonTest.parse_exact_size(b'not-a-valid-json')
        self.assertEqual(context_manager.exception.value, 'not-a-valid-json')

    def test_parse(self):
        header_field = FieldValueJsonTest.parse_exact_size(b'{"testTimeDelta": 1}')
        self.assertEqual(
            header_field,
            FieldValueJsonTest(datetime.timedelta(seconds=1))
        )
        self.assertEqual(
            header_field.string.value,  # pylint: disable=no-member
            attr.fields_dict(FieldValueJsonTest)['string'].default
        )
        self.assertEqual(
            header_field.number.value,  # pylint: disable=no-member
            attr.fields_dict(FieldValueJsonTest)['number'].default
        )

        parsed_header_field = FieldValueJsonTest.parse_exact_size(
            b'{"testTimeDelta": 1, "testString": "string", "testNumber": 1}'
        )
        header_field = FieldValueJsonTest(
            time_delta=datetime.timedelta(seconds=1),
            string='string',
            number=1
        )
        self.assertEqual(parsed_header_field, header_field)

        parsed_header_field = FieldValueJsonTest.parse_exact_size(b'{' + b', '.join([
            b'"testTimeDelta": 1',
            b'"testString": "string"',
            b'"testStringBase64": "ZGVmYXVsdA=="',
            b'"optional_string": "optional_string"',
            b'"testNumber": 1',
        ]) + b'}')
        header_field = FieldValueJsonTest(
            time_delta=datetime.timedelta(seconds=1),
            string='string',
            number=1
        )
        self.assertEqual(parsed_header_field, header_field)

    def test_compose(self):
        header_field = FieldValueJsonTest(datetime.timedelta(seconds=1))
        self.assertEqual(
            header_field.compose(),
            b'{' + b', '.join([
                b'"testTimeDelta": 1',
                b'"testString": "default"',
                b'"testUrl": "https://example.com"',
                b'"testStringBase64": "ZGVmYXVsdA=="',
                b'"testNumber": 0',
                b'"testPercent": 100',
            ]) + b'}'
        )


class TestFieldValueMultiple(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            FieldValueMultipleTest.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.value, None)

    def test_init_param_default(self):
        header_field = FieldValueMultipleTest(datetime.timedelta(1))
        self.assertEqual(header_field.option.value, False)
        self.assertEqual(header_field.number.value, 0)
        self.assertEqual(header_field.string.value, 'default')

    def test_init_param_convert(self):
        header_field_from_value = FieldValueMultipleTest(
            time_delta=datetime.timedelta(1),
        )
        header_field_from_component = FieldValueMultipleTest(
            time_delta=FieldValueComponentTimeDeltaTest(datetime.timedelta(1))
        )
        self.assertEqual(header_field_from_value, header_field_from_component)

    def test_parse(self):
        header_field = FieldValueMultipleTest.parse_exact_size(b'testTimeDelta=1')
        self.assertEqual(
            header_field,
            FieldValueMultipleTest(datetime.timedelta(seconds=1))
        )
        self.assertEqual(
            header_field.option.value,  # pylint: disable=no-member
            attr.fields_dict(FieldValueMultipleTest)['option'].default
        )
        self.assertEqual(
            header_field.string.value,  # pylint: disable=no-member
            attr.fields_dict(FieldValueMultipleTest)['string'].default
        )
        self.assertEqual(
            header_field.number.value,  # pylint: disable=no-member
            attr.fields_dict(FieldValueMultipleTest)['number'].default
        )

        parsed_header_field = FieldValueMultipleTest.parse_exact_size(
            b'testTimeDelta=1; testOption; testString=string; testNumber=1'
        )
        header_field = FieldValueMultipleTest(
            time_delta=datetime.timedelta(seconds=1),
            option=True,
            string='string',
            number=1
        )
        self.assertEqual(parsed_header_field, header_field)

        parsed_header_field = FieldValueMultipleTest.parse_exact_size(b'; '.join([
            b'testTimeDelta=1',
            b'testOption',
            b'testString=string',
            b'testStringBase64="ZGVmYXVsdA=="',
            b'testOptionalString=optional_string',
            b'testNumber=1',
        ]))
        header_field = FieldValueMultipleTest(
            time_delta=datetime.timedelta(seconds=1),
            option=True,
            string='string',
            optional_string='optional_string',
            number=1
        )
        self.assertEqual(parsed_header_field, header_field)

    def test_compose(self):
        header_field = FieldValueMultipleTest(datetime.timedelta(seconds=1))
        self.assertEqual(
            header_field.compose(),
            b'; '.join([
                b'testTimeDelta=1',
                b'testString=default',
                b'testUrl=https://example.com',
                b'testStringBase64="ZGVmYXVsdA=="',
                b'testNumber=0',
                b'testPercent=100',
            ])
        )

        header_field.option.value = True
        self.assertEqual(
            header_field.compose(),
            b'; '.join([
                b'testTimeDelta=1',
                b'testString=default',
                b'testUrl=https://example.com',
                b'testStringBase64="ZGVmYXVsdA=="',
                b'testNumber=0',
                b'testPercent=100',
                b'testOption',
            ])
        )


class TestFieldValueMultipleExtendable(unittest.TestCase):
    def test_parse(self):
        parsed_header_field = FieldValueMultipleExtendableTest.parse_exact_size(
            b'testTimeDelta=1; testExtension1=value1; testExtension2=value2'
        )
        header_field = FieldValueMultipleExtendableTest(
            time_delta=datetime.timedelta(seconds=1),
            extensions=NameValuePairListSemicolonSeparated(
                OrderedDict([('testExtension1', 'value1'), ('testExtension2', 'value2')])
            )
        )
        self.assertEqual(parsed_header_field, header_field)

    def test_compose(self):
        header_field = FieldValueMultipleExtendableTest(
            datetime.timedelta(seconds=1),
            extensions=NameValuePairListSemicolonSeparated(
                OrderedDict([('testExtension1', 'value1'), ('testExtension2', 'value2')])
            )
        )
        self.assertEqual(
            header_field.compose(),
            b'; '.join([
                b'testTimeDelta=1',
                b'testString=default',
                b'testUrl=https://example.com',
                b'testStringBase64="ZGVmYXVsdA=="',
                b'testNumber=0',
                b'testPercent=100',
                b'testExtension1=value1',
                b'testExtension2=value2',
            ])
        )


class TestFieldValueComponentKeyValue(unittest.TestCase):
    def test_parse_error(self):
        component = NameValuePair.parse_exact_size(b'name="')
        self.assertEqual(component.name, 'name')
        self.assertEqual(component.value, '')

        component = NameValuePair.parse_exact_size(b'name="value')
        self.assertEqual(component.name, 'name')
        self.assertEqual(component.value, 'value')

    def test_parse(self):
        component = NameValuePair.parse_exact_size(b'name=value')
        self.assertEqual(component.name, 'name')
        self.assertEqual(component.value, 'value')

        component = NameValuePair.parse_exact_size(b'name=')
        self.assertEqual(component.name, 'name')
        self.assertEqual(component.value, '')

    def test_parse_quoted(self):
        component = NameValuePair.parse_exact_size(b'name=""')
        self.assertEqual(component.name, 'name')
        self.assertEqual(component.value, '')

        component = NameValuePair.parse_exact_size(b'name="value"')
        self.assertEqual(component.name, 'name')
        self.assertEqual(component.value, 'value')

    def test_compose(self):
        self.assertEqual(NameValuePair('name', 'value').compose(), b'name=value')

    def test_compose_quoted(self):
        self.assertEqual(NameValuePair('name', 'value', quoted=True).compose(), b'name="value"')


class TestFieldValueDateTime(unittest.TestCase):
    def test_parse(self):
        http_header_field = FieldValueDateTime.parse_exact_size(b'Wed, 21 Oct 2015 07:28:00 GMT')
        self.assertEqual(
            http_header_field.value,
            datetime.datetime(2015, 10, 21, 7, 28, tzinfo=datetime.timezone.utc)
        )

        http_header_field = FieldValueDateTime.parse_exact_size(b'Wed, 21 Oct 2015 07:28:00 +01:00')
        self.assertEqual(
            http_header_field.value,
            datetime.datetime(2015, 10, 21, 7, 28, tzinfo=datetime.timezone(datetime.timedelta(hours=1)))
        )

    def test_compose(self):
        self.assertEqual(
            FieldValueDateTime(
                datetime.datetime(2015, 10, 21, 7, 28, tzinfo=datetime.timezone.utc)
            ).compose(),
            b'Wed, 21 Oct 2015 07:28:00 GMT'
        )


class TestFieldValueTimeDelta(unittest.TestCase):
    def test_parse(self):
        http_header_field = FieldValueTimeDeltaTest.parse_exact_size(b'86401')
        self.assertEqual(http_header_field.value, datetime.timedelta(days=1, seconds=1))

    def test_compose(self):
        self.assertEqual(
            FieldValueTimeDeltaTest(datetime.timedelta(days=1, seconds=1)).compose(),
            b'86401'
        )


class TestNameValuePairList(unittest.TestCase):
    _EMPTY_BYTES = b''
    _EMPTY = NameValuePairListSemicolonSeparated(OrderedDict([]))

    _OPTION_BYTES = b'option'
    _OPTION = NameValuePairListSemicolonSeparated(OrderedDict([('option', None)]))

    _KEY_VALUE_BYTES = b'key=value'
    _KEY_VALUE = NameValuePairListSemicolonSeparated(OrderedDict([('key', 'value')]))

    _KEY_VALUE_OPTION_BYTES = b'key=value; option'
    _KEY_VALUE_OPTION = NameValuePairListSemicolonSeparated(OrderedDict([('key', 'value'), ('option', None)]))

    _OPTION_KEY_VALUE_BYTES = b'option; key=value'
    _OPTION_KEY_VALUE = NameValuePairListSemicolonSeparated(OrderedDict([('option', None), ('key', 'value')]))

    def test_error(self):
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(b';;;option'),
            self._OPTION
        )
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(b'option;;;'),
            self._OPTION
        )

    def test_markdown(self):
        self.assertEqual(self._EMPTY.as_markdown(), '-')
        self.assertEqual(self._OPTION.as_markdown(), '* Option: n/a\n')
        self.assertEqual(self._KEY_VALUE.as_markdown(), '* Key: value\n')
        self.assertEqual(self._KEY_VALUE_OPTION.as_markdown(), '* Key: value\n* Option: n/a\n')
        self.assertEqual(self._OPTION_KEY_VALUE.as_markdown(), '* Option: n/a\n* Key: value\n')

    def test_parse_mixed_values(self):
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(self._EMPTY_BYTES),
            self._EMPTY
        )
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(self._OPTION_BYTES),
            self._OPTION
        )
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(self._KEY_VALUE_BYTES),
            self._KEY_VALUE
        )
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(self._KEY_VALUE_OPTION_BYTES),
            self._KEY_VALUE_OPTION
        )
        self.assertEqual(
            NameValuePairListSemicolonSeparated.parse_exact_size(self._OPTION_KEY_VALUE_BYTES),
            self._OPTION_KEY_VALUE
        )

    def test_compose_mixed_values(self):
        self.assertEqual(self._EMPTY.compose(), self._EMPTY_BYTES)
        self.assertEqual(self._OPTION.compose(), self._OPTION_BYTES)
        self.assertEqual(self._KEY_VALUE.compose(), self._KEY_VALUE_BYTES)
        self.assertEqual(self._KEY_VALUE_OPTION.compose(), self._KEY_VALUE_OPTION_BYTES)
        self.assertEqual(self._OPTION_KEY_VALUE.compose(), self._OPTION_KEY_VALUE_BYTES)

    def test_parse_single_option(self):
        header_field_value = NameValuePairListSemicolonSeparated.parse_exact_size(b'option')
        self.assertEqual(header_field_value.value, OrderedDict([('option', None), ]))

    def test_parse_muliple_options(self):
        header_field_value = NameValuePairListSemicolonSeparated.parse_exact_size(b'option1;option2;option3')
        self.assertEqual(
            header_field_value.value,
            OrderedDict([('option1', None), ('option2', None), ('option3', None), ])
        )

        header_field_value = NameValuePairListSemicolonSeparated.parse_exact_size(
            b' option1; \toption2;\t \toption3'
        )
        self.assertEqual(
            header_field_value.value,
            OrderedDict([('option1', None), ('option2', None), ('option3', None), ])
        )

    def test_compose_options(self):
        self.assertEqual(
            NameValuePairListSemicolonSeparated(OrderedDict([])).compose(),
            b''
        )
        self.assertEqual(self._OPTION.compose(), self._OPTION_BYTES)
        self.assertEqual(
            NameValuePairListSemicolonSeparated(
                OrderedDict([('option1', None), ('option2', None), ('option3', None), ])
            ).compose(),
            b'option1; option2; option3'
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import attr


from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.exception import NotEnoughData
from cryptoparser.tls.version import TlsVersion, TlsProtocolVersion
from cryptoparser.tls.subprotocol import TlsContentType
from cryptoparser.tls.subprotocol import SslMessageBase, SslMessageType, SslSubprotocolMessageParser


@attr.s
class TlsRecord(ParsableBase):
    HEADER_SIZE = 5

    fragment = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    protocol_version = attr.ib(
        default=TlsProtocolVersion(TlsVersion.TLS1),
        validator=attr.validators.instance_of(TlsProtocolVersion),
    )
    content_type = attr.ib(
        default=TlsContentType.HANDSHAKE,
        validator=attr.validators.instance_of(TlsContentType),
    )

    @classmethod
    def parse_header(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        try:
            parser.parse_numeric('content_type', 1, TlsContentType)
        except InvalidValue as e:
            raise InvalidValue(e.value, TlsContentType) from e
        parser.parse_parsable('protocol_version', TlsProtocolVersion)
        parser.parse_numeric('fragment_length', 2)

        return parser

    @classmethod
    def _parse(cls, parsable):
        parser = cls.parse_header(parsable)

        parser.parse_raw('fragment', parser['fragment_length'])

        return TlsRecord(
            content_type=parser['content_type'],
            protocol_version=parser['protocol_version'],
            fragment=parser['fragment'],
        ), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.content_type, 1)
        composer.compose_parsable(self.protocol_version)
        composer.compose_bytes(self.fragment, 2)

        return composer.composed_bytes


@attr.s
class SslRecord(ParsableBase):
    message = attr.ib(validator=attr.validators.instance_of(SslMessageBase))
    protocol_version = attr.ib(init=False, default=TlsVersion.SSL2, validator=attr.validators.in_(TlsVersion))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('record_length_0', 1)
        parser.parse_numeric('record_length_1', 1)
        if parser['record_length_0'] & 0x80:
            record_length = ((parser['record_length_0'] & 0x7f) * (2 ** 8)) + parser['record_length_1']
            padding_length = 0
        else:
            record_length = ((parser['record_length_0'] & 0x3f) * (2 ** 8)) + parser['record_length_1']
            parser.parse_numeric('padding_length', 1)
            padding_length = parser['padding_length']

        if record_length > parser.unparsed_length:
            raise NotEnoughData(record_length - parser.unparsed_length)

        try:
            parser.parse_numeric('message_type', 1, SslMessageType)
        except InvalidValue as e:
            raise InvalidValue(e.value, SslMessageType) from e

        parser.parse_variant('message', SslSubprotocolMessageParser(parser['message_type']))
        parser.parse_raw('padding', padding_length)

        return SslRecord(message=parser['message']), parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()
        message_type = self.message.get_message_type()
        body_composer.compose_numeric(message_type, 1)
        body_composer.compose_parsable(self.message)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(body_composer.composed_length | (2 ** 15), 2)

        return header_composer.composed_bytes + body_composer.composed_bytes

    @property
    def content_type(self):
        return self.message.get_message_type()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest
from unittest import mock

from cryptoparser.tls.grease import (
    TlsInvalidType,
    TlsInvalidTypeOneByte,
    TlsInvalidTypeParamsOneByte,
    TlsInvalidTypeParamsTwoByte,
    TlsInvalidTypeTwoByte,
)


class TestGrease(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(
            TlsInvalidTypeOneByte.parse_exact_size(b'\x2a').value,
            TlsInvalidTypeParamsOneByte(0x2a, TlsInvalidType.GREASE)
        )
        self.assertEqual(
            TlsInvalidTypeOneByte.parse_exact_size(b'\x2b').value,
            TlsInvalidTypeParamsOneByte(0x2b, TlsInvalidType.UNKNOWN)
        )

        self.assertEqual(
            TlsInvalidTypeTwoByte.parse_exact_size(b'\x2a\x2a').value,
            TlsInvalidTypeParamsTwoByte(0x2a2a, TlsInvalidType.GREASE)
        )
        self.assertEqual(
            TlsInvalidTypeTwoByte.parse_exact_size(b'\x2b\x2b').value,
            TlsInvalidTypeParamsTwoByte(0x2b2b, TlsInvalidType.UNKNOWN)
        )

    def test_compose(self):
        self.assertEqual(
            b'\x2a',
            TlsInvalidTypeOneByte(0x2a).compose()
        )
        self.assertEqual(
            b'\x2b',
            TlsInvalidTypeOneByte(0x2b).compose()
        )

        self.assertEqual(
            b'\x2a\x2a',
            TlsInvalidTypeTwoByte(0x2a2a).compose()
        )
        self.assertEqual(
            b'\x2b\x2b',
            TlsInvalidTypeTwoByte(0x2b2b).compose()
        )

    @mock.patch('random.choice')
    def test_random(self, mock_choice):
        mock_choice.side_effect = [0xabcd, ]
        self.assertEqual(TlsInvalidTypeTwoByte.from_random(), TlsInvalidTypeTwoByte(0xabcd))



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest


from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData

from cryptoparser.tls.record import TlsRecord
from cryptoparser.tls.subprotocol import TlsChangeCipherSpecMessage, TlsChangeCipherSpecType, TlsContentType
from cryptoparser.tls.version import TlsProtocolVersion, TlsVersion


class TestRecord(unittest.TestCase):
    def test_error(self):
        with self.assertRaisesRegex(InvalidValue, '0xff is not a valid TlsChangeCipherSpecType'):
            # pylint: disable=expression-not-assigned
            TlsChangeCipherSpecMessage.parse_exact_size(b'\xff')

        with self.assertRaises(NotEnoughData) as context_manager:
            # pylint: disable=expression-not-assigned
            TlsChangeCipherSpecMessage.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        self.assertEqual(
            TlsChangeCipherSpecMessage.parse_exact_size(b'\x01'),
            TlsChangeCipherSpecMessage(TlsChangeCipherSpecType.CHANGE_CIPHER_SPEC)
        )

    def test_compose(self):
        self.assertEqual(
            b'\x01',
            TlsChangeCipherSpecMessage(TlsChangeCipherSpecType.CHANGE_CIPHER_SPEC).compose()
        )

    def test_record(self):
        self.assertEqual(
            b'\x14\x03\x03\x00\x01\x01',
            TlsRecord(
                TlsChangeCipherSpecMessage().compose(),
                TlsProtocolVersion(TlsVersion.TLS1_2),
                TlsContentType.CHANGE_CIPHER_SPEC,
            ).compose()
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptoparser.tls.ciphersuite import TlsCipherSuite
from cryptoparser.tls.version import TlsProtocolVersion, TlsVersion


class TestTlsCipherSuite(unittest.TestCase):
    def test_str(self):
        for cipher_suite in filter(lambda cipher_suite: cipher_suite.value.iana_name, TlsCipherSuite):
            self.assertIn(cipher_suite.value.iana_name, str(cipher_suite.value))

        for cipher_suite in filter(lambda cipher_suite: cipher_suite.value.iana_name is None, TlsCipherSuite):
            self.assertIn(cipher_suite.name, str(cipher_suite.value))

        for cipher_suite in filter(lambda cipher_suite: cipher_suite.value.openssl_name, TlsCipherSuite):
            self.assertIn(cipher_suite.value.openssl_name, str(cipher_suite.value))

    def test_initial_version(self):
        self.assertEqual(
            TlsProtocolVersion(TlsCipherSuite.TLS_AES_128_GCM_SHA256.value.initial_version),
            TlsProtocolVersion(TlsVersion.TLS1_3_DRAFT_15)
        )
        self.assertEqual(
            TlsCipherSuite.TLS_AES_128_GCM_SHA256.value.initial_version,
            TlsVersion.TLS1_3_DRAFT_15
        )
        self.assertEqual(
            TlsProtocolVersion(TlsCipherSuite.TLS_RSA_WITH_NULL_SHA256.value.initial_version),
            TlsProtocolVersion(TlsVersion.TLS1_2)
        )
        self.assertEqual(
            TlsCipherSuite.TLS_RSA_WITH_NULL_SHA256.value.initial_version,
            TlsVersion.TLS1_2
        )



# SPDX-License-Identifier: MPL-2.0
"""IKEv1 payload parsers."""

import abc
import collections
import enum
import typing

import attr

from cryptodatahub.ike.algorithm import (
    Ikev1PayloadType,
    Ikev1ProtocolId,
    Ikev1EncryptionAlgorithm,
    Ikev1HashAlgorithm,
    Ikev1TransformId,
    Ikev1AttributeType,
    Ikev1Doi,
    Ikev1DiffieHellmanGroup,
    Ikev1AuthenticationMethod,
    Ikev1LifeType,
    Ikev1NotifyType,
)
from cryptoparser.common.base import VariantParsable
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.exception import NotEnoughData
from cryptoparser.ike.common import (
    DataAttributeBase,
    DataAttributeKeyLength,
    DataAttributeLength,
    DataAttributeTypeValueEnumCoded,
    DataAttributeFormat,
)


@attr.s
class Ikev1PayloadBase(ParsableBase):
    """Payload header parser, according to RFC2408.

    The generic payload header has the following structure:

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Next Payload  |    RESERVED   |         Payload Length        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :cvar HEADER_SIZE: Size of the header in bytes
    :ivar next_payload: Type of the next payload (1 byte)
    """
    HEADER_SIZE = 4

    next_payload: typing.Optional[Ikev1PayloadType] = attr.ib(
        init=False,
        default=Ikev1PayloadType.NONE,
        validator=attr.validators.optional(attr.validators.instance_of(Ikev1PayloadType))
    )

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_payload_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        """Parse payload header from bytes.

        :param parsable: Bytes to parse
        :type parsable: bytes
        :return: Tuple of (parsed header, number of bytes parsed)
        :rtype: tuple(PayloadBase, int)
        :raises NotEnoughData: If there are not enough bytes to parse
        """
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        parser.parse_numeric_enum_coded('next_payload', Ikev1PayloadType)
        parser.parse_numeric('reserved', 1)  # Skip reserved byte
        parser.parse_numeric('payload_length', 2)

        if parser.unparsed_length < parser['payload_length'] - cls.HEADER_SIZE:
            raise NotEnoughData(parser['payload_length'] - cls.HEADER_SIZE - parser.unparsed_length)

        return parser

    def compose_header(self, payload_length):
        """Compose payload header to bytes.

        :return: Composed header bytes
        :rtype: bytes
        """
        composer = ComposerBinary()
        composer.compose_numeric_enum_coded(self.next_payload)
        composer.compose_numeric(0, 1)  # Reserved byte
        composer.compose_numeric(self.HEADER_SIZE + payload_length, 2)

        return composer


@attr.s
class Ikev1AttributeAuthenticationMethod(DataAttributeTypeValueEnumCoded):
    """Authentication Method transform attribute (TLV format)."""

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.AUTHENTICATION_METHOD

    @classmethod
    def _get_enum_type(cls):
        return Ikev1AuthenticationMethod


@attr.s
class Ikev1AttributeDiffieHellmanGroup(DataAttributeTypeValueEnumCoded):
    """Diffie-Hellman Group transform attribute (TLV format)."""

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.GROUP_DESCRIPTION

    @classmethod
    def _get_enum_type(cls):
        return Ikev1DiffieHellmanGroup


@attr.s
class Ikev1AttributeKeyLength(DataAttributeKeyLength):
    """Key Length transform attribute (TV format)."""

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.KEY_LENGTH


@attr.s
class Ikev1AttributeEncryptionAlgorithm(DataAttributeTypeValueEnumCoded):
    """Encryption Algorithm transform attribute (TLV format)."""

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.ENCRYPTION_ALGORITHM

    @classmethod
    def _get_enum_type(cls):
        return Ikev1EncryptionAlgorithm


@attr.s
class Ikev1AttributeHashAlgorithm(DataAttributeTypeValueEnumCoded):
    """Hash Algorithm transform attribute (TLV format)."""

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.HASH_ALGORITHM

    @classmethod
    def _get_enum_type(cls):
        return Ikev1HashAlgorithm


@attr.s
class Ikev1AttributeLifeType(DataAttributeTypeValueEnumCoded):
    """Life Type transform attribute (TLV format)."""

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.LIFE_TYPE

    @classmethod
    def _get_enum_type(cls):
        return Ikev1LifeType


@attr.s
class Ikev1AttributeLifeDuration(DataAttributeLength):
    """Lifetime transform attribute (TLV format)."""

    @classmethod
    def _get_format(cls):
        return DataAttributeFormat.TYPE_LENGTH_VALUE

    @classmethod
    def get_type(cls):
        return Ikev1AttributeType.LIFE_DURATION

    @classmethod
    def _get_size(cls):
        return 4


@attr.s
class Ikev1PayloadTransform(Ikev1PayloadBase):
    """Transform Payload parser.

    The Transform Payload has the following format:

    .. code-block:: text

                            1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |    RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |  Transform #  |  Transform-Id |           RESERVED2           |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                        SA Attributes                          ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar transform_number: Transform number (1 byte)
    :ivar transform_id: Transform ID (1 byte)
    :ivar sa_attributes: Security Association attributes (variable length)
    """

    transform_id: Ikev1TransformId = attr.ib(validator=attr.validators.instance_of(Ikev1TransformId))
    attributes: typing.List[DataAttributeBase] = attr.ib(
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(DataAttributeBase),
        )
    )
    transform_number: typing.Optional[int] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int))
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.TRANSFORM

    def get_attribute_by_type(self, attribute_type: Ikev1AttributeType) -> DataAttributeBase:
        for attribute in self.attributes:
            if attribute.get_type() == attribute_type:
                return attribute
        raise KeyError(attribute_type)

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric('transform_number', 1)
        parser.parse_numeric_enum_coded('transform_id', Ikev1TransformId)
        parser.parse_numeric('reserved2', 2)  # Skip reserved2 field

        attributes = []
        parser_attributes = ParserBinary(parsable[parser.parsed_length:parser['payload_length']])
        while parser_attributes.unparsed_length > 0:
            parser_attributes.parse_parsable('attribute', Ikev1AttributeVariantServer)
            attribute = parser_attributes['attribute']
            attributes.append(attribute)

        payload = cls(
            transform_id=parser['transform_id'],
            attributes=attributes
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length + parser_attributes.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric(self.transform_number, 1)
        composer_payload.compose_numeric_enum_coded(self.transform_id)
        composer_payload.compose_numeric(0, 2)  # Reserved2 field

        for attribute in self.attributes:
            composer_payload.compose_parsable(attribute)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadProposal(Ikev1PayloadBase):
    """Proposal Payload parser.

    The Proposal Payload has the following format:

    .. code-block:: text

                          1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |    RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |  Proposal #   |  Protocol-Id  |    SPI Size   |# of Transforms|
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                        SPI (variable)                         |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar next_payload: Next payload type (1 byte)
    :ivar protocol_id: Protocol ID (1 byte)
    :ivar spi_size: Size of SPI in bytes (1 byte)
    :ivar transform_count: Number of transforms (1 byte)
    :ivar spi: Security Parameter Index (variable length)
    """

    protocol_id: Ikev1ProtocolId = attr.ib(validator=attr.validators.instance_of(Ikev1ProtocolId))
    transforms: typing.List[Ikev1PayloadTransform] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(Ikev1PayloadTransform),
    ))
    spi: bytes = attr.ib(default=bytes(), converter=bytes, validator=attr.validators.instance_of(bytes))
    next_payload: typing.Optional[Ikev1PayloadType] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(Ikev1PayloadType))
    )
    proposal_number: typing.Optional[int] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int))
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.PROPOSAL

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric('proposal_number', 1)
        parser.parse_numeric_enum_coded('protocol_id', Ikev1ProtocolId)
        parser.parse_numeric('spi_size', 1)
        parser.parse_numeric('transform_count', 1)
        parser.parse_raw('spi', parser['spi_size'])

        transforms = []
        for _ in range(parser['transform_count']):
            parser.parse_parsable('transform', Ikev1PayloadTransform)
            transforms.append(parser['transform'])

        payload = cls(
            protocol_id=parser['protocol_id'],
            spi=parser['spi'],
            transforms=transforms,
        )
        payload.next_payload = parser['next_payload']
        payload.proposal_number = parser['proposal_number']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric(self.proposal_number, 1)
        composer_payload.compose_numeric_enum_coded(self.protocol_id)
        composer_payload.compose_numeric(len(self.spi), 1)
        composer_payload.compose_numeric(len(self.transforms), 1)
        composer_payload.compose_raw(self.spi)

        for transform_number, transform in enumerate(self.transforms):
            transform.next_payload = (
                transform.get_payload_type()
                if transform_number < len(self.transforms) - 1
                else Ikev1PayloadType.NONE
            )
            transform.transform_number = transform_number + 1
            composer_payload.compose_parsable(transform)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


class Ikev1Situation(enum.IntFlag):
    """IKEv1 situation."""

    SIT_IDENTITY_ONLY = 1 << 0
    SIT_SECRECY = 1 << 1
    SIT_INTEGRITY = 1 << 2


@attr.s
class Ikev1PayloadSecurityAssociation(Ikev1PayloadBase):
    """Security Association payload parser.

    The Security Association payload has the following format:

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |    RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |              Domain of Interpretation  (DOI)                  |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                           Situation                           ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar doi: Domain of Interpretation (4 bytes)
    :ivar situation: Situation field (variable length)
    """

    doi: Ikev1Doi = attr.ib(validator=attr.validators.instance_of(Ikev1Doi))
    situation: typing.List[Ikev1Situation] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(Ikev1Situation),
    ))
    proposals: typing.List[Ikev1PayloadProposal] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(Ikev1PayloadProposal),
    ))

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.SECURITY_ASSOCIATION

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric_enum_coded('doi', Ikev1Doi)
        parser.parse_numeric_flags('situation', 4, Ikev1Situation)

        proposals = []
        parser_proposal = ParserBinary(parsable[parser.parsed_length:parser['payload_length']])
        while parser_proposal.unparsed_length > 0:
            parser_proposal.parse_parsable('proposal', Ikev1PayloadProposal)
            proposal = parser_proposal['proposal']
            proposals.append(proposal)

        payload = cls(
            doi=parser['doi'],
            situation=parser['situation'],
            proposals=proposals
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length + parser_proposal.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric_enum_coded(self.doi)
        composer_payload.compose_numeric_flags(self.situation, 4)

        for proposal_number, proposal in enumerate(self.proposals):
            proposal.next_payload = (
                proposal.get_payload_type()
                if proposal_number < len(self.proposals) - 1
                else Ikev1PayloadType.NONE
            )
            proposal.proposal_number = proposal_number + 1
            composer_payload.compose_parsable(proposal)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadKeyExchange(Ikev1PayloadBase):
    """Key Exchange payload parser.

    The Key Exchange payload has the following format:

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |    RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                           Key Exchange                          ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar next_payload: Next payload type (1 byte)
    :ivar key_exchange_data: Key exchange data (variable length)
    """

    key_exchange_data: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.KEY_EXCHANGE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_raw('key_exchange_data', parser['payload_length'] - cls.HEADER_SIZE)
        payload = cls(key_exchange_data=parser['key_exchange_data'])
        payload.next_payload = parser['next_payload']
        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.key_exchange_data)
        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadNonce(Ikev1PayloadBase):
    """Nonce payload parser.

    The Nonce payload has the following format:

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                           Nonce                               ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar nonce_data: Nonce data (variable length)
    """

    nonce_data: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.NONCE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_raw('nonce_data', parser['payload_length'] - cls.HEADER_SIZE)
        payload = cls(nonce_data=parser['nonce_data'])
        payload.next_payload = parser['next_payload']
        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.nonce_data)
        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadHash(Ikev1PayloadBase):
    """Hash payload parser.

    The Hash payload has the following format:

    .. code-block:: text

                          1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Next Payload  |   RESERVED    |         Payload Length        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                                                               |
        ~                           Hash Data                           ~
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar hash_data: Hash data (variable length)
    """

    hash_data: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.HASH

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_raw('hash_data', parser['payload_length'] - cls.HEADER_SIZE)
        payload = cls(hash_data=parser['hash_data'])
        payload.next_payload = parser['next_payload']
        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.hash_data)
        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadDoiProtocolSpiBase(Ikev1PayloadBase):
    """Base class for IKEv1 payloads with DOI, Protocol-Id, and SPI Size fields.

    Shared structure:

    .. code-block:: text

       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |              Domain of Interpretation  (DOI)                  |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |  Protocol-Id  |   SPI Size    |                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar doi: Domain of Interpretation (4 bytes)
    :ivar protocol_id: Protocol ID (1 byte)
    :ivar spi_size: Size of SPI in bytes (1 byte)
    """

    DOI_PROTOCOL_SPI_SIZE = 6

    doi: Ikev1Doi = attr.ib(validator=attr.validators.instance_of(Ikev1Doi))
    protocol_id: Ikev1ProtocolId = attr.ib(validator=attr.validators.instance_of(Ikev1ProtocolId))
    spi_size: int = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        parser = super()._parse_header(parsable)
        if parser.unparsed_length < cls.DOI_PROTOCOL_SPI_SIZE:
            raise NotEnoughData(cls.DOI_PROTOCOL_SPI_SIZE - parser.unparsed_length)

        parser.parse_numeric_enum_coded('doi', Ikev1Doi)
        parser.parse_numeric_enum_coded('protocol_id', Ikev1ProtocolId)
        parser.parse_numeric('spi_size', 1)

        return parser

    def _compose_doi_protocol_spi(self, composer):
        """Compose DOI, Protocol-Id, and SPI Size to composer."""
        composer.compose_numeric_enum_coded(self.doi)
        composer.compose_numeric_enum_coded(self.protocol_id)
        composer.compose_numeric(self.spi_size, 1)


@attr.s
class Ikev1PayloadNotification(Ikev1PayloadDoiProtocolSpiBase):
    """Notification payload parser.

    The Notification payload has the following format:

    .. code-block:: text

                          1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        ! Next Payload  !   RESERVED    !         Payload Length        !
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        !              Domain of Interpretation  (DOI)                  !
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        !  Protocol-ID  !   SPI Size    !      Notify Message Type      !
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        !                                                               !
        ~                Security Parameter Index (SPI)                 ~
        !                                                               !
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        !                                                               !
        ~                       Notification Data                       ~
        !                                                               !
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar notify_message_type: Notify message type (1 byte)
    :ivar spi: Security Parameter Index (variable length)
    :ivar notification_data: Notification data (variable length)
    """

    notify_type: Ikev1NotifyType = attr.ib(validator=attr.validators.instance_of(Ikev1NotifyType))
    spi: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))
    notification_data: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.NOTIFICATION

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_numeric_enum_coded('notify_type', Ikev1NotifyType)
        parser.parse_raw('spi', parser['spi_size'])
        parser.parse_raw('notification_data', parser['payload_length'] - parser.parsed_length)

        payload = cls(
            doi=parser['doi'],
            protocol_id=parser['protocol_id'],
            spi_size=parser['spi_size'],
            notify_type=parser['notify_type'],
            spi=parser['spi'],
            notification_data=parser['notification_data']
        )
        payload.next_payload = parser['next_payload']
        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        self._compose_doi_protocol_spi(composer_payload)
        composer_payload.compose_numeric_enum_coded(self.notify_type)
        composer_payload.compose_raw(self.spi)
        composer_payload.compose_raw(self.notification_data)

        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadDelete(Ikev1PayloadDoiProtocolSpiBase):
    """Delete payload parser.

    The Delete payload has the following format:

    .. code-block:: text

                          1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |   RESERVED    |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |              Domain of Interpretation  (DOI)                  |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |  Protocol-Id  |   SPI Size    |           # of SPIs           |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~               Security Parameter Index(es) (SPI)              ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar spis: Security Parameter Index(es) (variable length, list of spi_size bytes each)
    """

    spis: typing.Sequence[typing.Union[bytes, bytearray]] = attr.ib(
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of((bytes, bytearray)),
        )
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.DELETE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_numeric('spi_count', 2)

        spis = []
        for _ in range(parser['spi_count']):
            parser.parse_raw('spi', parser['spi_size'])
            spis.append(parser['spi'])

        payload = cls(
            doi=parser['doi'],
            protocol_id=parser['protocol_id'],
            spi_size=parser['spi_size'],
            spis=spis,
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        self._compose_doi_protocol_spi(composer_payload)
        composer_payload.compose_numeric(len(self.spis), 2)

        for spi in self.spis:
            composer_payload.compose_raw(spi)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadVendorId(Ikev1PayloadBase):
    """Vendor ID payload parser.

    The Vendor ID payload has the following format:

    .. code-block:: text

       1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      ! Next Payload  !C!  RESERVED   !         Payload Length        !
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      !                                                               !
      ~                        Vendor ID (VID)                        ~
      !                                                               !
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar vendor_id: Vendor ID (variable length)
    """

    vendor_id: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.VENDOR_ID

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_raw('vendor_id', parser['payload_length'] - cls.HEADER_SIZE)

        payload = cls(vendor_id=parser['vendor_id'])
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()

        composer_payload.compose_raw(self.vendor_id)
        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev1PayloadCertificateRequest(Ikev1PayloadBase):
    """Certificate Request payload parser.

    The Certificate Request payload has the following format:

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Next Payload  |C|  RESERVED   |         Payload Length        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Cert Encoding |                                               |
        +-+-+-+-+-+-+-+-+                                               +
        ~                       Certificate Data                        ~
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar cert_encoding: Certificate encoding type
    :ivar certificate_data: Certificate data
    """

    cert_encoding: int = attr.ib(validator=attr.validators.instance_of(int))
    certificate_data: typing.Union[bytes, bytearray] = attr.ib(
        validator=attr.validators.instance_of((bytes, bytearray))
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.CERTIFICATE_REQUEST

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_numeric('cert_encoding', 1)
        parser.parse_raw('certificate_data', parser['payload_length'] - cls.HEADER_SIZE - 1)

        payload = cls(
            cert_encoding=parser['cert_encoding'],
            certificate_data=parser['certificate_data']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric(self.cert_encoding, 1)
        composer_payload.compose_raw(self.certificate_data)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


IKEV1_PAYLOAD_CLASSES_BY_TYPE = {
    Ikev1PayloadType.SECURITY_ASSOCIATION: Ikev1PayloadSecurityAssociation,
    Ikev1PayloadType.KEY_EXCHANGE: Ikev1PayloadKeyExchange,
    Ikev1PayloadType.HASH: Ikev1PayloadHash,
    Ikev1PayloadType.NONCE: Ikev1PayloadNonce,
    Ikev1PayloadType.NOTIFICATION: Ikev1PayloadNotification,
    Ikev1PayloadType.DELETE: Ikev1PayloadDelete,
    Ikev1PayloadType.VENDOR_ID: Ikev1PayloadVendorId,
    Ikev1PayloadType.CERTIFICATE_REQUEST: Ikev1PayloadCertificateRequest,
}


class Ikev1AttributeVariantBase(VariantParsable):
    @classmethod
    @abc.abstractmethod
    def get_parsed_extensions(cls):
        raise NotImplementedError()

    @classmethod
    def _get_variants(cls):
        variants = cls.get_parsed_extensions()

        # variants.update([
        #     (extension_type, (Ikev1AttributeUnparsed, ))
        #     for extension_type in Ikev1AttributeType
        #     if extension_type not in variants
        # ])

        return variants


class Ikev1AttributeVariantServer(Ikev1AttributeVariantBase):
    @classmethod
    def get_parsed_extensions(cls):
        return collections.OrderedDict([
            (Ikev1AttributeType.ENCRYPTION_ALGORITHM, [Ikev1AttributeEncryptionAlgorithm, ]),
            (Ikev1AttributeType.HASH_ALGORITHM, [Ikev1AttributeHashAlgorithm, ]),
            (Ikev1AttributeType.LIFE_TYPE, [Ikev1AttributeLifeType, ]),
            (Ikev1AttributeType.KEY_LENGTH, [Ikev1AttributeKeyLength, ]),
            (Ikev1AttributeType.GROUP_DESCRIPTION, [Ikev1AttributeDiffieHellmanGroup, ]),
            (Ikev1AttributeType.AUTHENTICATION_METHOD, [Ikev1AttributeAuthenticationMethod, ]),
            (Ikev1AttributeType.LIFE_DURATION, [Ikev1AttributeLifeDuration, ]),
        ])



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptoparser.tls.subprotocol import TlsApplicationDataMessage


class TestRecord(unittest.TestCase):
    _APPLICATION_DATA_MESSAGE_BYTES = b'\x01\x02\x03\x04'

    def test_error(self):
        pass

    def test_parse(self):
        self.assertEqual(
            TlsApplicationDataMessage.parse_exact_size(self._APPLICATION_DATA_MESSAGE_BYTES),
            TlsApplicationDataMessage(data=self._APPLICATION_DATA_MESSAGE_BYTES)
        )

    def test_compose(self):
        self.assertEqual(
            self._APPLICATION_DATA_MESSAGE_BYTES,
            TlsApplicationDataMessage(data=self._APPLICATION_DATA_MESSAGE_BYTES).compose()
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest


from cryptodatahub.common.exception import InvalidValue
from cryptoparser.common.exception import NotEnoughData

from cryptoparser.tls.record import TlsRecord, SslRecord
from cryptoparser.tls.subprotocol import TlsContentType, TlsSubprotocolMessageBase
from cryptoparser.tls.version import TlsVersion, TlsProtocolVersion

from cryptoparser.tls.subprotocol import TlsAlertMessage, TlsAlertLevel, TlsAlertDescription
from cryptoparser.tls.subprotocol import SslErrorMessage, SslErrorType, SslMessageType


class TestTlsSubprotocolMessageBase(unittest.TestCase):
    def test_error(self):
        error_message = 'Can\'t instantiate abstract class TlsSubprotocolMessageBase'
        with self.assertRaisesRegex(BaseException, error_message) as context_manager:
            TlsSubprotocolMessageBase()  # pylint: disable=abstract-class-instantiated
        self.assertTrue(isinstance(context_manager.exception, (TypeError, AssertionError)))


class TestTlsRecord(unittest.TestCase):
    def setUp(self):
        self.test_message = TlsAlertMessage(
            level=TlsAlertLevel.FATAL,
            description=TlsAlertDescription.HANDSHAKE_FAILURE
        )
        self.test_record = TlsRecord(
            fragment=self.test_message.compose(),
            protocol_version=TlsProtocolVersion(TlsVersion.TLS1),
            content_type=TlsContentType.ALERT,
        )
        self.test_record_bytes = bytes(
            b'\x15' +      # type = ALERT
            b'\x03\x01' +  # version = TLS1
            b'\x00\x02' +  # length = 2
            b'\x02' +      # level = FATAL
            b'\x28' +      # description = HANDSHAKE_FAILURE
            b''
        )

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            TlsRecord.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, TlsRecord.HEADER_SIZE)

        with self.assertRaisesRegex(InvalidValue, '0xff is not a valid TlsContentType'):
            TlsRecord.parse_exact_size(
                b'\xff' +      # type = INVALID
                b'\x03\x03' +  # version = TLS 1.2
                b'\x00\x00' +  # length = 0
                b''
            )

        with self.assertRaises(NotEnoughData) as context_manager:
            TlsRecord.parse_exact_size(
                b'\x15' +      # type = alert
                b'\x03\x03' +  # version = TLS 1.2
                b'\x00\x01' +  # length = 1 (alert message is 2 bytes!)
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        TlsRecord.parse_exact_size(
            b'\x15' +      # type = alert
            b'\x03\x03' +  # version = TLS 1.2
            b'\x00\x02' +  # length = 2
            b'\x02\x28'
        )

        with self.assertRaises(NotEnoughData) as context_manager:
            TlsRecord.parse_exact_size(
                b'\x16' +      # type = handshake
                b'\x03\x03' +  # version = TLS 1.2
                b'\x00\x02' +  # length = 2 (handshake message is at least 4 bytes!)
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 2)

    def test_parse(self):
        record = TlsRecord.parse_exact_size(self.test_record_bytes)

        self.assertEqual(record.fragment, self.test_message.compose())
        self.assertEqual(record.protocol_version, TlsProtocolVersion(TlsVersion.TLS1))
        self.assertEqual(record.content_type, TlsContentType.ALERT)

    def test_compose(self):
        self.assertEqual(
            self.test_record.compose(),
            self.test_record_bytes
        )


class TestSslRecord(unittest.TestCase):
    def setUp(self):
        self.test_message = SslErrorMessage(
            error_type=SslErrorType.NO_CIPHER_ERROR
        )
        self.test_record = SslRecord(
            message=self.test_message
        )
        self.test_record_bytes = bytes(
            b'\x80\x03' +  # length = 3
            b'\x00' +      # message_type = ERROR
            b'\x00\x01' +  # error_type = NO_CIPHER_ERROR
            b''
        )

    def test_error(self):
        with self.assertRaisesRegex(InvalidValue, '0xff is not a valid SslMessageType'):
            SslRecord.parse_exact_size(
                b'\x80\x00' +  # length = 0
                b'\xff' +      # type = INVALID
                b''
            )

        with self.assertRaises(NotEnoughData) as context_manager:
            SslRecord.parse_exact_size(
                b'\x80\x03' +  # length = 3 (with length bytes)
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 3)

        with self.assertRaises(InvalidValue) as context_manager:
            SslRecord.parse_exact_size(
                b'\x80\x03' +  # length = 3
                b'\x00' +      # message_type = ERROR
                b'\x00\xff' +  # error_type = INVALID
                b''
            )

        with self.assertRaises(NotEnoughData) as context_manager:
            SslRecord.parse_exact_size(
                b'\x80\x03' +  # length = 3
                b'\x00' +      # type = ERROR
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 2)

        with self.assertRaises(NotEnoughData) as context_manager:
            SslRecord.parse_exact_size(
                b'\x81\x03' +  # length = 256 + 3
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 256 + 3 - 0)

        with self.assertRaises(NotEnoughData) as context_manager:
            SslRecord.parse_exact_size(
                b'\x01\x03\x01' +  # length = 256 + 3
                b''
            )
        self.assertEqual(context_manager.exception.bytes_needed, 256 + 3 - 0)

    def test_setter(self):
        record = SslRecord(self.test_message)

        record.message = self.test_message
        record.protocol_version = TlsVersion.SSL2

    def test_parse(self):
        record = SslRecord.parse_exact_size(self.test_record_bytes)

        self.assertEqual(
            record.message,
            self.test_message
        )
        self.assertEqual(record.content_type, SslMessageType.ERROR)

    def test_compose(self):
        self.assertEqual(
            self.test_record.compose(),
            self.test_record_bytes
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import functools

import attr

from cryptodatahub.common.grade import Grade, GradeableSimple
from cryptodatahub.tls.version import TlsVersion

from cryptoparser.common.base import TwoByteEnumParsable, ProtocolVersionBase
from cryptoparser.common.parse import ParserBinary, ComposerBinary


class TlsVersionFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsVersion

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s(order=False, eq=False, hash=True)
@functools.total_ordering
class TlsProtocolVersion(ProtocolVersionBase, GradeableSimple):
    version = attr.ib(validator=attr.validators.instance_of(TlsVersion))

    @property
    def grade(self):
        if self.version in (TlsVersion.TLS1_3, TlsVersion.TLS1_2):
            return Grade.SECURE
        if self.version in (TlsVersion.TLS1, TlsVersion.TLS1_1) or self.is_draft or self.is_google_experimental:
            return Grade.DEPRECATED
        if self.version in (TlsVersion.SSL2, TlsVersion.SSL3):
            return Grade.INSECURE

        raise NotImplementedError(self.version)

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_parsable('version', TlsVersionFactory)

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.major, 1)
        composer.compose_numeric(self.minor, 1)

        return composer.composed_bytes

    @property
    def major(self):
        return (self.version.value.code & 0xff00) >> 8

    @property
    def minor(self):
        return self.version.value.code & 0x00ff

    @property
    def is_draft(self):
        return self.major == 0x7f

    @property
    def is_google_experimental(self):
        return self.major == 0x7e

    def __eq__(self, other):
        return self.version.value.code == other.version.value.code

    def __lt__(self, other):
        if self.major == other.major:
            return self.minor < other.minor
        if self.is_draft:
            return other.version == TlsVersion.TLS1_3
        if other.is_draft:
            return self.version != TlsVersion.TLS1_3

        return self.major < other.major

    @property
    def identifier(self):
        return self.version.name.lower()

    def __str__(self):
        if self.is_draft:
            return f'TLS 1.3 Draft {self.minor}'
        if self.is_google_experimental:
            return f'TLS 1.3 Google Experiment {self.minor}'
        if self.version == TlsVersion.SSL3:
            return 'SSL 3.0'
        if self.version == TlsVersion.SSL2:
            return 'SSL 2.0'

        return f'TLS 1.{self.minor - 1}'

    def _as_markdown(self, level):
        return self._markdown_result(str(self), level)

    def _asdict(self):
        return self.identifier



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections
import unittest

from cryptodatahub.ike.algorithm import Ikev2NotifyType, Ikev2ProtocolId

from cryptoparser.common.exception import NotEnoughData
from cryptoparser.ike.ikev2 import (
    Ikev2PayloadFlags,
    Ikev2PayloadType,
    Ikev2PayloadNotifyUnparsed,
    Ikev2NotifyPayloadCookie,
    Ikev2NotifyPayloadVariantResponder
)

from .classes import Ikev2PayloadNotifyBaseTest, Ikev2PayloadNotifyNoDataTest


class TestIkev2PayloadNotifyBase(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.protocol_id = Ikev2ProtocolId.IKE
        self.notify_type = Ikev2NotifyType.AUTHENTICATION_FAILED
        self.spi = b'\x00\x01\x02\x03'
        self.test_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

        self.notify_payload_minimal = Ikev2PayloadNotifyBaseTest(
            flags=set(),
            protocol_id=self.protocol_id,
            notify_type=self.notify_type,
            spi=bytes(),
            test_data=b''
        )
        self.notify_payload_minimal.next_payload = Ikev2PayloadType.NONE

        self.notify_payload_with_data = Ikev2PayloadNotifyBaseTest(
            flags={Ikev2PayloadFlags.CRITICAL},
            protocol_id=self.protocol_id,
            notify_type=self.notify_type,
            spi=self.spi,
            test_data=self.test_data
        )
        self.notify_payload_with_data.next_payload = Ikev2PayloadType.KE

        self.notify_dict_minimal = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE
            ('flags', b'\x00'),  # No flags
            ('payload_length', b'\x00\x08'),  # 4 bytes header + 4 bytes notify header + 0 bytes data
            ('protocol_id', b'\x01'),  # IKE
            ('spi_size', b'\x00'),  # 0 bytes SPI
            ('notify_type', b'\x00\x18'),  # AUTHENTICATION_FAILED (0x0018)
        ])
        self.notify_bytes_minimal = b''.join(self.notify_dict_minimal.values())

        self.notify_dict_with_data = collections.OrderedDict([
            ('next_payload', b'\x22'),  # KE
            ('flags', b'\x80'),  # CRITICAL
            ('payload_length', b'\x00\x1c'),  # 4 bytes header + 4 bytes notify header + 4 bytes SPI + 16 bytes data
            ('protocol_id', b'\x01'),  # IKE
            ('spi_size', b'\x04'),  # 4 bytes SPI
            ('notify_type', b'\x00\x18'),  # AUTHENTICATION_FAILED (0x0018)
            ('spi', self.spi),  # SPI data
            ('test_data', self.test_data),  # Notification data
        ])
        self.notify_bytes_with_data = b''.join(self.notify_dict_with_data.values())

    def test_get_payload_type(self):
        self.assertEqual(Ikev2PayloadNotifyBaseTest.get_payload_type(), Ikev2PayloadType.NOTIFY)

    def test_parse(self):
        parsed_minimal: Ikev2PayloadNotifyBaseTest = Ikev2PayloadNotifyBaseTest.parse_exact_size(
            self.notify_bytes_minimal
        )
        self.assertEqual(parsed_minimal.flags, set())
        self.assertEqual(parsed_minimal.next_payload, Ikev2PayloadType.NONE)
        self.assertEqual(parsed_minimal.protocol_id, self.protocol_id)
        self.assertEqual(parsed_minimal.type, self.notify_type)
        self.assertEqual(parsed_minimal.spi, bytes())
        self.assertEqual(parsed_minimal.test_data, b'')

        parsed_with_data: Ikev2PayloadNotifyBaseTest = Ikev2PayloadNotifyBaseTest.parse_exact_size(
            self.notify_bytes_with_data
        )
        self.assertEqual(parsed_with_data.flags, {Ikev2PayloadFlags.CRITICAL})
        self.assertEqual(parsed_with_data.next_payload, Ikev2PayloadType.KE)
        self.assertEqual(parsed_with_data.protocol_id, self.protocol_id)
        self.assertEqual(parsed_with_data.type, self.notify_type)
        self.assertEqual(parsed_with_data.spi, self.spi)
        self.assertEqual(parsed_with_data.test_data, self.test_data)

    def test_compose(self):
        composed_minimal = self.notify_payload_minimal.compose()
        self.assertEqual(composed_minimal, self.notify_bytes_minimal)

        composed_with_data = self.notify_payload_with_data.compose()
        self.assertEqual(composed_with_data, self.notify_bytes_with_data)

    def test_round_trip(self):
        composed_minimal = self.notify_payload_minimal.compose()
        parsed_minimal: Ikev2PayloadNotifyBaseTest = Ikev2PayloadNotifyBaseTest.parse_exact_size(composed_minimal)

        self.assertEqual(parsed_minimal.protocol_id, self.notify_payload_minimal.protocol_id)
        self.assertEqual(parsed_minimal.type, self.notify_payload_minimal.type)
        self.assertEqual(parsed_minimal.spi, self.notify_payload_minimal.spi)
        self.assertEqual(parsed_minimal.test_data, self.notify_payload_minimal.test_data)
        self.assertEqual(parsed_minimal.flags, self.notify_payload_minimal.flags)
        self.assertEqual(parsed_minimal.next_payload, self.notify_payload_minimal.next_payload)

        composed_with_data = self.notify_payload_with_data.compose()
        parsed_with_data: Ikev2PayloadNotifyBaseTest = Ikev2PayloadNotifyBaseTest.parse_exact_size(composed_with_data)

        self.assertEqual(parsed_with_data.protocol_id, self.notify_payload_with_data.protocol_id)
        self.assertEqual(parsed_with_data.type, self.notify_payload_with_data.type)
        self.assertEqual(parsed_with_data.spi, self.notify_payload_with_data.spi)
        self.assertEqual(parsed_with_data.test_data, self.notify_payload_with_data.test_data)
        self.assertEqual(parsed_with_data.flags, self.notify_payload_with_data.flags)
        self.assertEqual(parsed_with_data.next_payload, self.notify_payload_with_data.next_payload)

    def test_error_parse_not_enough_data(self):
        incomplete_data = self.notify_bytes_minimal[:-2]

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev2PayloadNotifyBaseTest.parse_exact_size(incomplete_data)
        self.assertEqual(context_manager.exception.bytes_needed, 2)

    def test_error_invalid_protocol_id(self):
        with self.assertRaises(TypeError):
            Ikev2PayloadNotifyBaseTest(
                flags=set(),
                protocol_id="invalid",
                notify_type=self.notify_type,
                spi=bytes(),
                test_data=b''
            )

    def test_error_invalid_notify_type(self):
        with self.assertRaises(TypeError):
            Ikev2PayloadNotifyBaseTest(
                flags=set(),
                protocol_id=self.protocol_id,
                notify_type="invalid",
                spi=bytes(),
                test_data=b''
            )

    def test_error_invalid_spi(self):
        with self.assertRaises(TypeError):
            Ikev2PayloadNotifyBaseTest(
                flags=set(),
                protocol_id=self.protocol_id,
                notify_type=self.notify_type,
                spi=None,
                test_data=b''
            )


class TestIkev2PayloadNotifyNoData(unittest.TestCase):
    def setUp(self):
        self.wrong_notify_dict = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE
            ('flags', b'\x00'),  # No flags
            ('payload_length', b'\x00\x08'),  # 4 bytes header + 4 bytes notify header
            ('protocol_id', b'\x01'),  # IKE
            ('spi_size', b'\x00'),  # 0 bytes SPI
            ('notify_type', b'\x00\x01'),  # UNSUPPORTED_CRITICAL_PAYLOAD (not AUTHENTICATION_FAILED)
        ])
        self.wrong_notify_bytes = b''.join(self.wrong_notify_dict.values())

    def test_get_message_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev2PayloadNotifyNoDataTest._get_message_type(), Ikev2NotifyType.AUTHENTICATION_FAILED)

    def test_error_invalid_notify_type(self):
        with self.assertRaises(Exception):  # InvalidType from the implementation
            Ikev2PayloadNotifyNoDataTest.parse_exact_size(self.wrong_notify_bytes)


class TestIkev2PayloadNotifyUnparsed(unittest.TestCase):
    def setUp(self):
        self.protocol_id = Ikev2ProtocolId.IKE
        self.notify_type = Ikev2NotifyType.INVALID_SELECTORS  # Different from AUTHENTICATION_FAILED
        self.spi = b'\x00\x01\x02\x03'
        self.data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

        self.notify_payload_with_data = Ikev2PayloadNotifyUnparsed(
            flags={Ikev2PayloadFlags.CRITICAL},
            protocol_id=self.protocol_id,
            type=self.notify_type,
            spi=self.spi,
            data=self.data
        )
        self.notify_payload_with_data.next_payload = Ikev2PayloadType.KE

    def test_any_notify_type_support(self):
        different_notify_types = [
            Ikev2NotifyType.AUTHENTICATION_FAILED,
            Ikev2NotifyType.INVALID_SELECTORS,
            Ikev2NotifyType.UNSUPPORTED_CRITICAL_PAYLOAD,
        ]

        for notify_type in different_notify_types:
            payload = Ikev2PayloadNotifyUnparsed(
                flags=set(),
                protocol_id=self.protocol_id,
                type=notify_type,
                spi=bytes(),
                data=b'\x00\x01\x02\x03'
            )
            self.assertEqual(payload.type, notify_type)
            self.assertEqual(payload.data, b'\x00\x01\x02\x03')

    def test_raw_data_storage(self):
        payload = Ikev2PayloadNotifyUnparsed(
            flags=set(),
            protocol_id=self.protocol_id,
            type=self.notify_type,
            spi=bytes(),
            data=self.data
        )
        self.assertEqual(payload.data, self.data)

        different_data = b'\xff\xfe\xfd\xfc'
        payload_2 = Ikev2PayloadNotifyUnparsed(
            flags=set(),
            protocol_id=self.protocol_id,
            type=self.notify_type,
            spi=bytes(),
            data=different_data
        )
        self.assertEqual(payload_2.data, different_data)

    def test_round_trip_data_preservation(self):
        payload_no_spi = Ikev2PayloadNotifyUnparsed(
            flags=set(),
            protocol_id=self.protocol_id,
            type=self.notify_type,
            spi=bytes(),
            data=self.data
        )
        payload_no_spi.next_payload = Ikev2PayloadType.NONE

        composed_bytes = payload_no_spi.compose()
        parsed_payload: Ikev2PayloadNotifyUnparsed = Ikev2PayloadNotifyUnparsed.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_payload.data, payload_no_spi.data)  # pylint: disable=no-member
        self.assertEqual(parsed_payload.type, payload_no_spi.type)
        self.assertEqual(parsed_payload.spi, payload_no_spi.spi)

    def test_parse_with_spi(self):
        data_with_spi = b'\x00\x01\x02\x03\x04\x05\x06\x07'
        payload_with_spi = Ikev2PayloadNotifyUnparsed(
            flags=set(),
            protocol_id=self.protocol_id,
            type=self.notify_type,
            spi=self.spi,
            data=data_with_spi
        )
        payload_with_spi.next_payload = Ikev2PayloadType.NONE
        composed_bytes = payload_with_spi.compose()
        parsed_payload: Ikev2PayloadNotifyUnparsed = Ikev2PayloadNotifyUnparsed.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_payload.protocol_id, self.protocol_id)
        self.assertEqual(parsed_payload.type, self.notify_type)
        self.assertEqual(parsed_payload.spi, self.spi)
        self.assertEqual(parsed_payload.data, data_with_spi)  # pylint: disable=no-member
        self.assertEqual(parsed_payload.flags, set())
        self.assertEqual(parsed_payload.next_payload, Ikev2PayloadType.NONE)


class TestIkev2NotifyPayloadCookie(unittest.TestCase):
    def setUp(self):
        self.protocol_id = Ikev2ProtocolId.IKE
        self.cookie_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

        self.cookie_payload = Ikev2NotifyPayloadCookie(
            flags=set(),
            protocol_id=self.protocol_id,
            type=Ikev2NotifyType.COOKIE,
            spi=bytes(),
            cookie=self.cookie_data
        )
        self.cookie_payload.next_payload = Ikev2PayloadType.NONE

    def test_get_message_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev2NotifyPayloadCookie._get_message_type(), Ikev2NotifyType.COOKIE)

    def test_cookie_data_storage(self):
        payload = Ikev2NotifyPayloadCookie(
            flags=set(),
            protocol_id=self.protocol_id,
            type=Ikev2NotifyType.COOKIE,
            spi=bytes(),
            cookie=self.cookie_data
        )
        self.assertEqual(payload.cookie, self.cookie_data)

        different_cookie = b'\xff\xfe\xfd\xfc\xfb\xfa'
        payload_2 = Ikev2NotifyPayloadCookie(
            flags=set(),
            protocol_id=self.protocol_id,
            type=Ikev2NotifyType.COOKIE,
            spi=bytes(),
            cookie=different_cookie
        )
        self.assertEqual(payload_2.cookie, different_cookie)

    def test_round_trip_cookie_preservation(self):
        composed_bytes = self.cookie_payload.compose()
        parsed_payload: Ikev2NotifyPayloadCookie = Ikev2NotifyPayloadCookie.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_payload.cookie, self.cookie_payload.cookie)  # pylint: disable=no-member
        self.assertEqual(parsed_payload.type, self.cookie_payload.type)
        self.assertEqual(parsed_payload.spi, self.cookie_payload.spi)


class TestIkev2NotifyPayloadVariantResponder(unittest.TestCase):
    def setUp(self):
        self.protocol_id = Ikev2ProtocolId.IKE

    def test_parse_other_notify_type(self):
        other_notify = Ikev2PayloadNotifyUnparsed(
            flags=set(),
            protocol_id=self.protocol_id,
            type=Ikev2NotifyType.AUTHENTICATION_FAILED,
            spi=bytes(),
            data=b'\x00\x01\x02\x03'
        )
        other_notify.next_payload = Ikev2PayloadType.NONE
        composed_bytes = other_notify.compose()

        parsed_payload = Ikev2NotifyPayloadVariantResponder.parse_exact_size(composed_bytes)
        self.assertIsInstance(parsed_payload, Ikev2PayloadNotifyUnparsed)
        self.assertEqual(parsed_payload.type, Ikev2NotifyType.AUTHENTICATION_FAILED)
        self.assertEqual(parsed_payload.data, b'\x00\x01\x02\x03')  # pylint: disable=no-member

    def test_compose(self):
        cookie_data = b'\x00\x01\x02\x03'
        cookie_payload = Ikev2NotifyPayloadCookie(
            flags=set(),
            protocol_id=self.protocol_id,
            type=Ikev2NotifyType.COOKIE,
            spi=bytes(),
            cookie=cookie_data
        )
        cookie_payload.next_payload = Ikev2PayloadType.NONE

        variant_parsable = Ikev2NotifyPayloadVariantResponder(variant=cookie_payload)
        composed_bytes = variant_parsable.compose()

        self.assertEqual(composed_bytes, cookie_payload.compose())

    def test_round_trip(self):
        cookie_data = b'\x00\x01\x02\x03\x04\x05'
        cookie_payload = Ikev2NotifyPayloadCookie(
            flags=set(),
            protocol_id=self.protocol_id,
            type=Ikev2NotifyType.COOKIE,
            spi=bytes(),
            cookie=cookie_data
        )
        cookie_payload.next_payload = Ikev2PayloadType.NONE

        variant_parsable = Ikev2NotifyPayloadVariantResponder(variant=cookie_payload)
        composed_bytes = variant_parsable.compose()
        parsed_payload = Ikev2NotifyPayloadVariantResponder.parse_exact_size(composed_bytes)

        self.assertIsInstance(parsed_payload, Ikev2NotifyPayloadCookie)
        self.assertEqual(parsed_payload.cookie, cookie_data)  # pylint: disable=no-member
        self.assertEqual(parsed_payload.type, cookie_payload.type)
        self.assertEqual(parsed_payload.spi, cookie_payload.spi)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import base64
import collections
import datetime
import unittest

from cryptodatahub.common.algorithm import Authentication, KeyExchange, NamedGroup
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.key import (
    PublicKey,
    PublicKeyParamsDsa,
    PublicKeyParamsEcdsa,
    PublicKeyParamsEddsa,
    PublicKeyParamsRsa,
)

from cryptodatahub.dnsrec.algorithm import DnsSecAlgorithm, DnsSecDigestType, DnsRrType

from cryptoparser.common.exception import NotEnoughData
from cryptoparser.dnsrec.record import (
    DnsNameUncompressed,
    DnsRecordDnskey,
    DnsRecordDs,
    DnsRecordMx,
    DnsRecordRrsig,
    DnsRecordTxt,
    DnsRrTypePrivate,
    DnsSecFlag,
    DnsSecProtocol,
)


class TestDnsRecordDnskey(unittest.TestCase):
    def setUp(self):
        self.header_bytes = (
            b'\x01\x00' +   # flags: DNS_ZONE_KEY
            b'\x03' +       # version
            b''
        )

    def test_error_inconsistent_algorithm(self):
        public_key_rsa = PublicKey.from_params(PublicKeyParamsRsa(
            public_exponent=2 ** 2048 - 1,
            modulus=2 ** 1024 - 1,
        ))

        with self.assertRaises(InvalidValue) as context_manager:
            DnsRecordDnskey(
                flags=[DnsSecFlag.DNS_ZONE_KEY],
                algorithm=DnsSecAlgorithm.DH,
                key=public_key_rsa,
                protocol=DnsSecProtocol.V3,
            )
        self.assertEqual(context_manager.exception.value, KeyExchange.DH)

        with self.assertRaises(InvalidValue) as context_manager:
            DnsRecordDnskey(
                flags=[DnsSecFlag.DNS_ZONE_KEY],
                algorithm=DnsSecAlgorithm.ECCGOST,
                key=public_key_rsa,
                protocol=DnsSecProtocol.V3,
            )
        self.assertEqual(context_manager.exception.value, Authentication.GOST_R3410_01)

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            DnsRecordDnskey.parse_exact_size(self.header_bytes)

        self.assertEqual(
            context_manager.exception.bytes_needed,
            DnsRecordDnskey.HEADER_SIZE - len(self.header_bytes)
        )

    def test_key_tag(self):
        record_bytes = self.header_bytes + (
            b'\x01' +            # algorithm: RSAMD5
            b'\x03' +            # exponent_length: 3
            b'\x01\x00\x01' +    # exponent: 65537
            124 * b'\x00' +      # modulus
            b'\x11\x22\x44\x88'
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.RSAMD5)
        self.assertEqual(dns_record.key_tag, 0x2244)

        record_bytes = self.header_bytes + (
            b'\x05' +          # algorithm: RSASHA1
            b'\x03' +          # exponent_length: 3
            b'\x01\x00\x01' +  # exponent: 65537
            128 * b'\xff' +    # modulus
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.RSASHA1)
        self.assertEqual(dns_record.key_tag, 1799)

    def test_asdict(self):
        self.assertEqual(
            DnsRecordDnskey(
                flags=[DnsSecFlag.DNS_ZONE_KEY],
                algorithm=DnsSecAlgorithm.RSASHA1,
                key=PublicKey.from_params(PublicKeyParamsRsa(
                    public_exponent=2 ** 2048 - 1,
                    modulus=2 ** 1024 - 1,
                )),
                protocol=DnsSecProtocol.V3,
            )._asdict(),
            collections.OrderedDict([
                ('key_tag', 1540),
                ('flags', [DnsSecFlag.DNS_ZONE_KEY]),
                ('algorithm', DnsSecAlgorithm.RSASHA1),
                ('key', PublicKey.from_params(PublicKeyParamsRsa(
                    public_exponent=2 ** 2048 - 1,
                    modulus=2 ** 1024 - 1,
                ))),
                ('protocol', DnsSecProtocol.V3),
            ])

        )

    def test_parse_rsa_key(self):
        record_bytes = self.header_bytes + (
            b'\x05' +          # algorithm: RSASHA1
            b'\x03' +          # exponent_length: 3
            b'\x01\x00\x01' +  # exponent: 65537
            128 * b'\xff' +    # modulus
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.RSASHA1)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsRsa(
            public_exponent=65537,
            modulus=2 ** 1024 - 1,
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

        record_bytes = self.header_bytes + (
            b'\x05' +          # algorithm: RSASHA1
            b'\x00' +          # exponent length extender mark
            b'\x01\x00' +      # exponent_length: 256
            256 * b'\xff' +    # exponent
            128 * b'\xff' +    # modulus
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.RSASHA1)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsRsa(
            public_exponent=2 ** 2048 - 1,
            modulus=2 ** 1024 - 1,
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

    def test_parse_dsa_key(self):
        record_bytes = self.header_bytes + (
            b'\x03' +                               # algorithm: DSA
            b'\x08' +                               # key size parameter
            20 * b'\xff' +                          # q
            b'\x80' + (1024 // 8 - 1) * b'\x00' +   # p
            b'\x40' + (1024 // 8 - 1) * b'\x00' +   # g
            b'\x20' + (1024 // 8 - 1) * b'\x00' +   # y
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.DSA)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsDsa(
            prime=2 ** 1023,
            generator=2 ** 1022,
            order=2 ** 160 - 1,
            public_key_value=2 ** 1021,
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

    def test_parse_ecdsa_key(self):
        record_bytes = self.header_bytes + (
            b'\x0c' +                               # algorithm: ECCGOST
            b'\x80' + (256 // 8 - 1) * b'\x00' +     # point_x
            b'\x40' + (256 // 8 - 1) * b'\x00' +     # point_y
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.ECCGOST)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsEcdsa(
            named_group=NamedGroup.GC256B,
            point_x=2 ** 255,
            point_y=2 ** 254,
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

        record_bytes = self.header_bytes + (
            b'\x0d' +                               # algorithm: ECDSAP256SHA256
            b'\x80' + (256 // 8 - 1) * b'\x00' +     # point_x
            b'\x40' + (256 // 8 - 1) * b'\x00' +     # point_y
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.ECDSAP256SHA256)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsEcdsa(
            named_group=NamedGroup.SECP256K1,
            point_x=2 ** 255,
            point_y=2 ** 254,
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

        record_bytes = self.header_bytes + (
            b'\x0e' +                               # algorithm: ECDSAP384SHA384
            b'\x80' + (384 // 8 - 1) * b'\x00' +     # point_x
            b'\x40' + (384 // 8 - 1) * b'\x00' +     # point_y
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.ECDSAP384SHA384)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsEcdsa(
            named_group=NamedGroup.SECP384R1,
            point_x=2 ** 383,
            point_y=2 ** 382,
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

    def test_parse_eddsa_key(self):
        record_bytes = self.header_bytes + (
            b'\x0f' +          # algorithm: ED25519
            32 * b'\xff' +     # key_data
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.ED25519)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsEddsa(
            curve_type=NamedGroup.CURVE25519,
            key_data=32 * b'\xff',
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

        record_bytes = self.header_bytes + (
            b'\x10' +          # algorithm: ED448
            56 * b'\xff' +     # key_data
            b''
        )
        dns_record = DnsRecordDnskey.parse_exact_size(record_bytes)
        self.assertEqual(dns_record.algorithm, DnsSecAlgorithm.ED448)
        self.assertEqual(dns_record.key, PublicKey.from_params(PublicKeyParamsEddsa(
            curve_type=NamedGroup.CURVE448,
            key_data=56 * b'\xff',
        )))
        self.assertEqual(dns_record.compose(), record_bytes)

    def test_real(self):
        # RFC 4034 Section 5.4
        public_key_bytes = base64.b64decode(
            'AQOeiiR0GOMYkDshWoSKz9Xz'
            'fwJr1AYtsmx3TGkJaNXVbfi/'
            '2pHm822aJ5iI9BMzNXxeYCmZ'
            'DRD99WYwYqUSdjMmmAphXdvx'
            'egXd/M5+X7OrzKBaMbCVdFLU'
            'Uh6DhweJBjEVv5f2wwjM9Xzc'
            'nOf+EPbtG9DMBmADjFDc2w/r'
            'ljwvFw=='
        )
        public_key = DnsRecordDnskey.parse_key(public_key_bytes, DnsSecAlgorithm.RSASHA1)
        self.assertEqual(DnsRecordDnskey.compose_key(public_key), public_key_bytes)

        dns_record = DnsRecordDnskey(
            flags=[DnsSecFlag.DNS_ZONE_KEY],
            algorithm=DnsSecAlgorithm.RSASHA1,
            key=public_key,
            protocol=DnsSecProtocol.V3,
        )
        self.assertEqual(dns_record.key_tag, 60485)

        # RFC 4034 Section 2.3
        public_key_bytes = base64.b64decode(
            'AQPSKmynfzW4kyBv015MUG2DeIQ3'
            'Cbl+BBZH4b/0PY1kxkmvHjcZc8no'
            'kfzj31GajIQKY+5CptLr3buXA10h'
            'WqTkF7H6RfoRqXQeogmMHfpftf6z'
            'Mv1LyBUgia7za6ZEzOJBOztyvhjL'
            '742iU/TpPSEDhm2SNKLijfUppn1U'
            'aNvv4w=='
        )
        public_key = DnsRecordDnskey.parse_key(public_key_bytes, DnsSecAlgorithm.RSASHA1)
        self.assertEqual(DnsRecordDnskey.compose_key(public_key), public_key_bytes)

        dns_record = DnsRecordDnskey(
            flags=[DnsSecFlag.DNS_ZONE_KEY],
            algorithm=DnsSecAlgorithm.RSASHA1,
            key=public_key,
            protocol=DnsSecProtocol.V3,
        )
        self.assertEqual(dns_record.key_tag, 2642)

        # RFC 5702 Section 6.1
        public_key_bytes = base64.b64decode(
            'AwEAAcFcGsaxxdgiuuGmCkVI'
            'my4h99CqT7jwY3pexPGcnUFtR2Fh36BponcwtkZ4cAgtvd4Qs8P'
            'kxUdp6p/DlUmObdk='
        )
        public_key = DnsRecordDnskey.parse_key(public_key_bytes, DnsSecAlgorithm.RSASHA256)
        self.assertEqual(DnsRecordDnskey.compose_key(public_key), public_key_bytes)

        dns_record = DnsRecordDnskey(
            flags=[DnsSecFlag.DNS_ZONE_KEY],
            algorithm=DnsSecAlgorithm.RSASHA256,
            key=public_key,
            protocol=DnsSecProtocol.V3,
        )
        self.assertEqual(dns_record.key_tag, 9033)

        # RFC 5702 Section 6.2
        public_key_bytes = base64.b64decode(
            'AwEAAdHoNTOW+et86KuJOWRD'
            'p1pndvwb6Y83nSVXXyLA3DLroROUkN6X0O6pnWnjJQujX/AyhqFD'
            'xj13tOnD9u/1kTg7cV6rklMrZDtJCQ5PCl/D7QNPsgVsMu1J2Q8g'
            'pMpztNFLpPBz1bWXjDtaR7ZQBlZ3PFY12ZTSncorffcGmhOL'
        )
        public_key = DnsRecordDnskey.parse_key(public_key_bytes, DnsSecAlgorithm.RSASHA512)
        self.assertEqual(DnsRecordDnskey.compose_key(public_key), public_key_bytes)

        dns_record = DnsRecordDnskey(
            flags=[DnsSecFlag.DNS_ZONE_KEY],
            algorithm=DnsSecAlgorithm.RSASHA512,
            key=public_key,
            protocol=DnsSecProtocol.V3,
        )
        self.assertEqual(dns_record.key_tag, 3740)

        # RFC 5933 Section 2.2
        public_key_bytes = base64.b64decode(
            'aRS/DcPWGQj2wVJydT8EcAVoC0kXn5pDVm2I'
            'MvDDPXeD32dsSKcmq8KNVzigjL4OXZTV+t/6'
            'w4X1gpNrZiC01g=='
        )
        public_key = DnsRecordDnskey.parse_key(public_key_bytes, DnsSecAlgorithm.ECCGOST)
        self.assertEqual(DnsRecordDnskey.compose_key(public_key), public_key_bytes)

        dns_record = DnsRecordDnskey(
            flags=[DnsSecFlag.DNS_ZONE_KEY],
            algorithm=DnsSecAlgorithm.ECCGOST,
            key=public_key,
            protocol=DnsSecProtocol.V3,
        )
        self.assertEqual(dns_record.key_tag, 59732)


class TestDnsRecordDs(unittest.TestCase):
    def setUp(self):
        self.record_bytes = bytes(
            b'\x00\x01' +        # key_tag: 1
            b'\x01' +            # algorithm: RSAMD5
            b'\x02' +            # digest_type: SHA_256
            32 * b'\xff' +       # digest
            b''
        )
        self.record = DnsRecordDs(
            key_tag=1,
            algorithm=DnsSecAlgorithm.RSAMD5,
            digest_type=DnsSecDigestType.SHA_256,
            digest=32 * b'\xff',
        )

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            DnsRecordDs.parse_exact_size(b'\x00')

        self.assertEqual(
            context_manager.exception.bytes_needed,
            DnsRecordDs.HEADER_SIZE - 1
        )

    def test_parse(self):
        self.assertEqual(DnsRecordDs.parse_exact_size(self.record_bytes), self.record)

    def test_compose(self):
        self.assertEqual(self.record.compose(), self.record_bytes)


class TestDnsNameUncompressed(unittest.TestCase):
    def setUp(self):
        self.label_empty_bytes = b'\x00'
        self.label_empty_name = DnsNameUncompressed([])
        self.label_single_bytes = b'\x01a\x00'
        self.label_single_name = DnsNameUncompressed(['a'])
        self.label_multiple_bytes = b'\x01a\x02bb\x03ccc\x00'
        self.label_multiple_name = DnsNameUncompressed(['a', 'bb', 'ccc'])

    def test_error_convert_invalid_value(self):
        with self.assertRaises(InvalidValue) as context_manager:
            DnsNameUncompressed.convert(None)

        self.assertEqual(context_manager.exception.value, None)

    def test_parse(self):
        self.assertEqual(DnsNameUncompressed.parse_exact_size(self.label_empty_bytes), self.label_empty_name)
        self.assertEqual(DnsNameUncompressed.parse_exact_size(self.label_single_bytes), self.label_single_name)
        self.assertEqual(DnsNameUncompressed.parse_exact_size(self.label_multiple_bytes), self.label_multiple_name)

    def test_compose(self):
        self.assertEqual(self.label_empty_name.compose(), self.label_empty_bytes)
        self.assertEqual(self.label_single_name.compose(), self.label_single_bytes)
        self.assertEqual(self.label_multiple_name.compose(), self.label_multiple_bytes)

    def test_convert(self):
        self.assertEqual(DnsNameUncompressed.convert(self.label_empty_name), self.label_empty_name)
        self.assertEqual(DnsNameUncompressed.convert(self.label_single_name), self.label_single_name)
        self.assertEqual(DnsNameUncompressed.convert(self.label_multiple_name), self.label_multiple_name)

        self.assertEqual(DnsNameUncompressed.convert(''), self.label_empty_name)
        self.assertEqual(DnsNameUncompressed.convert('a'), self.label_single_name)
        self.assertEqual(DnsNameUncompressed.convert('a.bb.ccc'), self.label_multiple_name)

    def test_str(self):
        self.assertEqual(str(self.label_empty_name), '')
        self.assertEqual(str(self.label_single_name), 'a')
        self.assertEqual(str(self.label_multiple_name), 'a.bb.ccc')

    def test_as_markdown(self):
        self.assertEqual(self.label_empty_name.as_markdown(), '')
        self.assertEqual(self.label_single_name.as_markdown(), 'a')
        self.assertEqual(self.label_multiple_name.as_markdown(), 'a.bb.ccc')


class TestDnsRecordRrsig(unittest.TestCase):
    def setUp(self):
        self.record_bytes = bytes(
            b'\x00\x01' +          # type_covered: A
            b'\x01' +              # algorithm: RSAMD5
            b'\x03' +              # labels
            b'\x00\x00\x0e\x10' +  # original_ttl: 3600
            b'\x00\x00\x00\x01' +  # signature_expiration
            b'\x00\x00\x00\x02' +  # signature_inception
            b'\xab\xcd' +          # key_tag
            b'\x06signer\x00' +    # signers_name
            32 * b'\xff' +         # signature
            b''
        )
        self.record = DnsRecordRrsig(
            type_covered=DnsRrType.A,
            algorithm=DnsSecAlgorithm.RSAMD5,
            labels=3,
            original_ttl=3600,
            signature_expiration=datetime.datetime(1970, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc),
            signature_inception=datetime.datetime(1970, 1, 1, 0, 0, 2, tzinfo=datetime.timezone.utc),
            key_tag=0xabcd,
            signers_name='signer',
            signature=32 * b'\xff',
        )

        self.record_private_type_bytes = bytes(
            b'\xff\xfe' +          # type_covered: A
            b'\x01' +              # algorithm: RSAMD5
            b'\x03' +              # labels
            b'\x00\x00\x0e\x10' +  # original_ttl: 3600
            b'\x00\x00\x00\x01' +  # signature_expiration
            b'\x00\x00\x00\x02' +  # signature_inception
            b'\xab\xcd' +          # key_tag
            b'\x06signer\x00' +    # signers_name
            32 * b'\xff' +         # signature
            b''
        )
        self.record_private_types = DnsRecordRrsig(
            type_covered=DnsRrTypePrivate(0xfffe),
            algorithm=DnsSecAlgorithm.RSAMD5,
            labels=3,
            original_ttl=3600,
            signature_expiration=datetime.datetime(1970, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc),
            signature_inception=datetime.datetime(1970, 1, 1, 0, 0, 2, tzinfo=datetime.timezone.utc),
            key_tag=0xabcd,
            signers_name='signer',
            signature=32 * b'\xff',
        )

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            DnsRecordRrsig.parse_exact_size(b'\x00')

        self.assertEqual(
            context_manager.exception.bytes_needed,
            DnsRecordRrsig.HEADER_SIZE - 1
        )

    def test_parse(self):
        self.assertEqual(DnsRecordRrsig.parse_exact_size(self.record_bytes), self.record)
        self.assertEqual(DnsRecordRrsig.parse_exact_size(self.record_private_type_bytes), self.record_private_types)

    def test_compose(self):
        self.assertEqual(self.record.compose(), self.record_bytes)
        self.assertEqual(self.record_private_types.compose(), self.record_private_type_bytes)


class TestDnsRecordMx(unittest.TestCase):
    def setUp(self):
        self.record_bytes = bytes(
            b'\x00\x01' +          # priority: 1
            b'\x08exchange\x00' +  # exchange
            b''
        )
        self.record = DnsRecordMx(
            priority=1,
            exchange='exchange',
        )

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            DnsRecordMx.parse_exact_size(b'\x00')

        self.assertEqual(
            context_manager.exception.bytes_needed,
            DnsRecordMx.HEADER_SIZE - 1
        )

    def test_parse(self):
        self.assertEqual(DnsRecordMx.parse_exact_size(self.record_bytes), self.record)

    def test_compose(self):
        self.assertEqual(self.record.compose(), self.record_bytes)


class TestDnsRecordTxt(unittest.TestCase):
    def setUp(self):
        self.record_bytes_single = bytes(
            b'\x05' +              # length: 5
            b'value' +
            b''
        )
        self.record_single = DnsRecordTxt(value='value')

        self.record_bytes_multiple = bytes(
            b'\x06' +              # length: 6
            b'value1' +
            b'\x06' +              # length: 6
            b'value2' +
            b''
        )
        self.record_multiple = DnsRecordTxt(value='value1value2')

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            DnsRecordTxt.parse_exact_size(b'')

        self.assertEqual(
            context_manager.exception.bytes_needed,
            DnsRecordTxt.HEADER_SIZE
        )

    def test_parse(self):
        self.assertEqual(DnsRecordTxt.parse_exact_size(self.record_bytes_single), self.record_single)
        self.assertEqual(DnsRecordTxt.parse_exact_size(self.record_bytes_multiple), self.record_multiple)

    def test_compose(self):
        self.assertEqual(self.record_single.compose(), self.record_bytes_single)
